import os
import json
import shutil
import yaml as YAML
from typing import List
from .figurl_to_html import _parse_figurl_url, _get_dependent_uris_recursive, _parse_sha1_uri, _download_view
import kachery_cloud as kcl
import requests

thisdir = os.path.dirname(os.path.realpath(__file__))

def figurl_doc_to_html(*,
    doc_fname: str,
    output_dir: str
):
    if os.path.exists(output_dir):
        print(f'File or folder already exists: {output_dir}')
    os.makedirs(f'{output_dir}/views')
    os.makedirs(f'{output_dir}/sha1')
    os.makedirs(f'{output_dir}/images')

    figurls: List[str] = _get_figurls_from_doc(doc_fname)
    images: List[dict] = _get_images_from_doc(doc_fname)
    view_uris: List[str] = []
    data_uris: List[str] = []
    for url in figurls:
        F = _parse_figurl_url(url)
        if F.v not in view_uris:
            view_uris.append(F.v)
        aa = _get_dependent_uris_recursive(uri=F.d)
        for uu in aa:
            if uu not in data_uris:
                data_uris.append(uu)
    for data_uri in data_uris:
        sha1 = _parse_sha1_uri(data_uri)
        dst_fname = f'{output_dir}/sha1/{sha1}'
        print(f'Writing {dst_fname}')
        kcl.load_file(data_uri, dest=dst_fname)
    for view_uri in view_uris:
        pp = _path_for_view_uri(view_uri)
        view_dir = f'{output_dir}/views/{pp}'
        os.makedirs(view_dir)
        _download_view(view_uri, output_dir=view_dir)
    for image in images:
        image_name: str = image['name']
        image_url: str = image['url']
        if '/' in image_name:
            dd = f'{output_dir}/images/{os.path.dirname(image_name)}'
            if not os.path.exists(dd):
                os.makedirs(dd)
        dest_path = f'{output_dir}/images/{image_name}'
        print(f'Downloading file to: {dest_path}')
        _download_file(image_url, dest=dest_path)
    
    pp = f'{output_dir}/index.md'
    print(f'Writing {pp}')
    shutil.copyfile(doc_fname, pp)

    print('Downloading doc-figurl')
    _download_view('gs://figurl/doc-figurl', output_dir=output_dir)

    print(f'Writing {output_dir}/figurl.json')
    figurl_json = {
        'figurlToHtml': True
    }
    with open(f'{output_dir}/figurl.json', 'w') as f:
        json.dump(figurl_json, f)

def _path_for_view_uri(uri: str):
    return uri.replace('://', '/')

def _get_figurls_from_doc(doc_fname: str):
    ret: List[str] = []
    with open(doc_fname, 'r') as f:
        txt = f.read()
    lines = txt.split('\n')
    for line in lines:
        if line.startswith('https://figurl.org/f'):
            figurl = line.strip()
            if figurl not in ret:
                ret.append(figurl)
    return ret

def _get_images_from_doc(doc_fname: str):
    ret: List[dict] = []
    with open(doc_fname, 'r') as f:
        txt = f.read()
    lines = txt.split('\n')
    for i in range(len(lines)):
        line = lines[i]
        if line.startswith('!['):
            a = _get_comment_parameters(lines[i + 1:])
            if a is None:
                raise Exception(f'Comment parameters required after image at line {i + 1}')
            if 'name' not in a:
                raise Exception(f'Parameter "name" required after image at line {i + 1}')
            name: str = a['name']
            try:
                i1 = line.index('(')
                i2 = line.index(')')
            except:
                raise Exception(f'Unexpected format of image at line {i + 1}')
            url = line[i1 + 1:i2]
            ret.append({'name': name, 'url': url})
    return ret

def _get_comment_parameters(lines: List[str]):
    if (lines[0].strip() != '<!--'):
        return None
    yaml_lines: List[str] = []
    i = 1
    while (i < len(lines)):
        line = lines[i]
        if line.strip() == '-->':
            break
        yaml_lines.append(line)
        i += 1
    yaml = '\n'.join(yaml_lines)
    return YAML.safe_load(yaml)

def _download_file(url: str, *, dest: str):
    # see: https://stackoverflow.com/questions/16694907/download-large-file-in-python-with-requests
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(dest, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                # specifying chunk size is important, because the default apparently is 1
                # if the source is chunk-encoded, perhaps chunk_size should be None?
                if chunk: 
                    f.write(chunk)


import os
from typing import List, Union
import kachery_cloud as kcl


def generate_doc_bibliography(*, file_path: str, bibliography_path: str, csl_path: Union[str, None]):
    with open(file_path, 'r') as f:
        txt = f.read()
    lines = txt.split('\n')
    bib_comment_text = '<!--bibliography-->' 
    if bib_comment_text in lines:
        ind = lines.index(bib_comment_text)
        body_lines = lines[:ind]
    else:
        print(f'WARNING: No "{bib_comment_text}" line found in file.')
        return

    bib_keys: List[str] = []
    for line in body_lines:
        a = _get_bib_keys(line)
        for b in a:
            if b not in bib_keys:
                bib_keys.append(b)
    
    bib_lines: List[str] = []
    for i, bib_key in enumerate(bib_keys):
        print(f'Generating bibliography item for {bib_key}')
        txt = _get_bib_item_text(key=bib_key, bibliography_path=bibliography_path, csl_path=csl_path)
        if txt:
            bib_lines.append(
                f'\\[@{bib_key}\\] {txt}'
            )
            if i < len(bib_keys) - 1:
                bib_lines.append('')
        else:
            print(f'WARNING: Unable to find {bib_key}')

    new_lines = body_lines + ['<!--bibliography-->', ''] + bib_lines
    new_txt = '\n'.join(new_lines)
    with open(file_path, 'w') as f:
        f.write(new_txt)

def _get_bib_keys(line: str):
    keys: List[str] = []
    def process_separated_part(x: str):
        if x.startswith('@'):
            keys.append(x[1:])
        return x
    def process_bracketed_part(x: str):
        _handleSeparatedParts(x, [',', ';'], process_separated_part)
        return x
    _handleBracketedParts(line, process_bracketed_part)
    return keys

def _handleBracketedParts(a: str, handler: callable) -> str:
    parts1: List[str] = a.split('[')
    parts2: List[str] = []
    for i, part1 in enumerate(parts1):
        if i > 0:
            jj = part1.find(']')
            if jj > 0:
                parts2.append(handler(part1[:jj]) + part1[jj:])
            else:
                parts2.append(part1)
        else:
            parts2.append(part1)
    return '['.join(parts2)

def _handleSeparatedParts(a: str, delimiters: List[str], handler: callable) -> str:
    for d in delimiters:
        b = a.split(d)
        if len(b) > 1:
            return d.join([_handleSeparatedParts(x, delimiters, handler) for x in b])
    if a.startswith(' '):
        return ' ' + handler(a[1:])
    else:
        return handler(a)

def _get_bib_item_text(*, key: str, bibliography_path: str, csl_path: Union[str, None]):
    # should we call this a hack? yes.
    with kcl.TemporaryDirectory() as tmpdir:
        with open(f'{tmpdir}/input.md', 'w') as f:
            f.write(f'''
{key}: [@{key}]

<!---->

''')
        opts = [
            f'--bibliography {bibliography_path}'
        ]
        if csl_path is not None:
            opts.append(
                f'--csl {csl_path}'
            )
        os.system(f'pandoc -t markdown_strict {" ".join(opts)} -o {tmpdir}/output.md {tmpdir}/input.md')

        with open(f'{tmpdir}/output.md', 'r') as f:
            a = f.read()
        lines = a.split('\n')
        ind = lines.index('<!---->')
        txt = '\n'.join(lines[ind + 1:]).strip()
        prefixes_to_remove = '\\[1\\] '
        for p in prefixes_to_remove:
            if txt.startswith(p):
                txt = txt[len(p):]
        return txt


import os
import shutil
import requests
import json
import time
from typing import List
from dataclasses import dataclass
import kachery_cloud as kcl

thisdir = os.path.dirname(os.path.realpath(__file__))

def figurl_to_html(*,
    url: str,
    output_dir: str
):
    F = _parse_figurl_url(url)
    if not F.d:
        raise Exception('No d parameter')
    if not F.v:
        raise Exception('No v parameter')
    if os.path.exists(output_dir):
        print(f'File or folder already exists: {output_dir}')
    os.makedirs(f'{output_dir}/sha1')
    os.makedirs(f'{output_dir}/view')

    _download_view(F.v, output_dir=output_dir + '/view')

    config_fname = f'{output_dir}/figurl.json'
    print(f'Writing {config_fname}')
    with open(config_fname, 'w') as f:
        json.dump({
            'd': F.d,
            'v': F.v,
            'label': F.label,
            'zone': F.zone,
            'url': url,
            'timestampCreated': time.time()
        }, f, indent=4)

    dependent_uris = _get_dependent_uris_recursive(uri=F.d)
    for uri in dependent_uris:
        if uri.startswith('sha1://'):
            sha1 = _parse_sha1_uri(uri)
            dst_fname = f'{output_dir}/sha1/{sha1}'
            print(f'Writing {dst_fname}')
            kcl.load_file(uri, dest=dst_fname)
        else:
            print(f'Depends on: {uri}')
    
    dstpath = f'{output_dir}/index.html'
    print(f'Writing {dstpath}')
    shutil.copyfile(f'{thisdir}/templates/figurl_to_html_index.html', dstpath)

def _get_dependent_uris_recursive(*, uri: str):
    ret: List[str] = []
    ret.append(uri)
    a = kcl.load_text(uri)
    try:
        b = json.loads(a)
    except:
        b = None
    if b is not None:
        x = _get_dependent_uris_from_obj_recursive(obj=b)
        for c in x:
            if not c in ret:
                ret.append(c)
    return ret

def _get_dependent_uris_from_obj_recursive(*, obj):
    ret: List[str] = []
    if isinstance(obj, str):
        if obj.startswith('sha1://') or obj.startswith('zenodo://') or obj.startswith('zenodo-sandbox://'):
            if not obj in ret:
                aa = _get_dependent_uris_recursive(uri=obj)
                for bb in aa:
                    if not bb in ret:
                        ret.append(bb)
    elif isinstance(obj, list) or isinstance(obj, tuple):
        for c in obj:
            a = _get_dependent_uris_from_obj_recursive(obj=c)
            for b in a:
                if not b in ret:
                    ret.append(b)
    elif isinstance(obj, dict):
        for k, v in obj.items():
            a = _get_dependent_uris_from_obj_recursive(obj=v)
            for b in a:
                if not b in ret:
                    ret.append(b)
    else:
        pass
    return ret

@dataclass
class FigurlUrl:
    d: str
    v: str
    s: str
    label: str
    zone: str

def _parse_figurl_url(uri: str):
    ind = uri.index('?')
    q = uri[ind + 1:]
    a = q.split('&')
    query = {}
    for b in a:
        x = b.split('=')
        if len(x) == 2:
            query[x[0]] = x[1]
    return FigurlUrl(
        d=query.get('d', ''),
        v=query.get('v', ''),
        s=query.get('s', ''),
        label=query.get('label', ''),
        zone=query.get('zone', '')
    )

def _parse_sha1_uri(uri: str):
    a = uri.split('?')[0].split('/')
    sha1 = a[2]
    return sha1

def _http_get(url: str):
    r = requests.get(url)
    if r.status_code == 200:
        return r.text
    else:
        raise Exception(f'Error getting {url}: {r.status_code}')

def _download_view(view_uri: str, *, output_dir: str):
    if view_uri.startswith('sha1://') or view_uri.startswith('zenodo://') or view_uri.startswith('zenodo-sandbox://'):
        view_fname = f'{output_dir}/index.html'
        print(f'Writing {view_fname}')
        kcl.load_file(view_uri, dest=view_fname)
    elif view_uri.startswith('gs://'):
        p = view_uri[len("gs://"):]
        url = f'https://storage.googleapis.com/{p}'
        index_html = _http_get(f'{url}/index.html')
        _write_file(f'{output_dir}/index.html', index_html)
        manifest_json = _http_get(f'{url}/manifest.json')
        _write_file(f'{output_dir}/manifest.json', manifest_json)
        asset_manifest_json = _http_get(f'{url}/asset-manifest.json')
        _write_file(f'{output_dir}/asset-manifest.json', asset_manifest_json)
        asset_manifest = json.loads(asset_manifest_json)
        for k, f in asset_manifest['files'].items():
            if f.startswith('/'):
                f = f[1:]
            url_src = f'{url}/{f}'
            pp_dst = f'{output_dir}/{f}'
            os.makedirs(os.path.dirname(pp_dst), exist_ok=True)
            txt = _http_get(url_src)
            _write_file(pp_dst, txt)

def _write_file(fname: str, content: str):
    print(f'Writing {fname}')
    with open(fname, 'w') as f:
        f.write(content)


from typing import List
import os
import requests
import hashlib


class DepositionFile:
    def __init__(self, f: dict) -> None:
        self.checksum: str = f['checksum']
        self.filename: str = f['filename']
        self.filesize: int = f['filesize']
        self.id: str = f['id']
        self.links = f['links']

def upload_to_zenodo(*, folder: str, deposition_id: str, sandbox: bool):
    env_name = 'ZENODO_SANDBOX_ACCESS_TOKEN' if sandbox else 'ZENODO_ACCESS_TOKEN'
    try:
        access_token = os.environ[env_name]
    except:
        raise Exception(f'Environment variable not set: {env_name}')
    zenodo_base_url = 'https://sandbox.zenodo.org' if sandbox else 'https://zenodo.org'
    deposition = _http_get_json(
        f'{zenodo_base_url}/api/deposit/depositions/{deposition_id}',
        params={'access_token': access_token}
    )
    bucket_url = deposition["links"]["bucket"]
    deposition_files = [DepositionFile(f) for f in deposition['files']]
    filenames_in_deposition = set()
    for f in deposition_files:
        to_delete = False
        if not os.path.exists(f'{folder}/{f.filename}'):
            to_delete = True
        elif not _file_matches(f'{folder}/{f.filename}', filesize=f.filesize, checksum=f.checksum):
            to_delete = True
        if to_delete:
            print(f'Deleting from deposition: {f.filename}')
            _http_delete(
                f'{zenodo_base_url}/api/deposit/depositions/{deposition_id}/files/{f.filename}',
                params={'access_token': access_token}
            )
        else:
            filenames_in_deposition.add(f.filename)
    filenames = _get_all_filenames_in_dir(folder)
    for filename in filenames:
        if filename in filenames_in_deposition:
            print(f'Already in deposition: {filename}')
        else:
            print(f'Uploading {filename}')
            with open(f'{folder}/{filename}', 'rb') as ff:
                _http_put(
                    f'{bucket_url}/{filename}',
                    params={'access_token': access_token},
                    data=ff
                )
    print(f'{zenodo_base_url}/deposit/{deposition_id}')

def _http_get_json(url: str, *, params):
    resp = requests.get(
        url,
        params=params
    )
    if resp.status_code != 200:
        raise Exception(f'Error getting: {url} ({resp.status_code})')
    return resp.json()

def _http_delete(url: str, *, params):
    resp = requests.delete(url, params=params)
    if resp.status_code != 204:
        raise Exception(f'Error deleting: {url} ({resp.status_code})')

def _http_put(url: str, *, params, data):
    resp = requests.put(url, params=params, data=data)
    if resp.status_code != 200:
        raise Exception(f'Error uploading: {url} ({resp.status_code})')

def _file_matches(path: str, *, filesize: int, checksum: str):
    if os.path.getsize(path) != filesize:
        return False
    if os.path.getsize(path) > 1024 * 10:
        print(f'Computing md5 checksum of {path}')
    md5 = _compute_md5_of_file(path)
    if md5 != checksum:
        return False
    return True

def _compute_md5_of_file(path: str):
    with open(path, 'rb') as f:
        data = f.read()
        md5 = hashlib.md5(data).hexdigest()
        return md5

def _get_all_filenames_in_dir(path: str):
    ret: List[str] = []
    for f in os.listdir(path):
        pp = f'{path}/{f}'
        if os.path.isfile(pp):
            ret.append(f)
    for f in os.listdir(path):
        pp = f'{path}/{f}'
        if os.path.isdir(pp):
            for b in _get_all_filenames_in_dir(pp):
                ret.append(f'{f}/{b}')
    return ret


from typing import Union
import click
from .figurl_to_html import figurl_to_html
from .figurl_doc_to_html import figurl_doc_to_html
from .upload_to_zenodo import upload_to_zenodo as do_upload_to_zenodo
from .view_doc import view_doc as do_view_doc
from .generate_doc_bibliography import generate_doc_bibliography as do_generate_doc_bibliography


@click.group(help="Isa command-line client")
def cli():
    pass

@click.command(help="Create static HTML folder for a figurl figure")
@click.option('--input', required=True, help="The figurl URL for the figure")
@click.option('--output', required=True, help="The output folder")
def create_figure(input: str, output: str):
    figurl_to_html(url=input, output_dir=output)

@click.command(help="Create static HTML folder for a figurl document")
@click.option('--input', required=True, help='The .md document file')
@click.option('--output', required=True, help="The output folder")
def create_doc(input: str, output: str):
    figurl_doc_to_html(doc_fname=input, output_dir=output)

@click.command(help="Upload a figure or doc folder to Zenodo")
@click.option('--folder', required=True, help="The local folder to upload")
@click.option('--deposition-id', required=True, help="The ID of the Zenodo deposition")
@click.option('--sandbox', is_flag=True, default=False, help='Use sandbox Zenodo site')
def upload_to_zenodo(folder: str, deposition_id: str, sandbox: bool):
    do_upload_to_zenodo(folder=folder, deposition_id=deposition_id, sandbox=sandbox)

@click.command(help="View a figurl markdown document locally")
@click.argument('file')
@click.option('--port', default=8000, help="The local port to use")
@click.option('--allow-origin', default='https://doc.figurl.org', help="The URL of doc.figurl.org to allow")
def view_doc(file: str, port: int, allow_origin: str):
    do_view_doc(file_path=file, port=port, allow_origin=allow_origin)

@click.command(help="Generate doc bibliography")
@click.argument('file')
@click.option('--bibliography', required=True, help="Path to .bib file")
@click.option('--csl', help="Path to csl file")
def generate_doc_bibliography(file: str, bibliography: str, csl: Union[str, None]):
    do_generate_doc_bibliography(file_path=file, bibliography_path=bibliography, csl_path=csl)

cli.add_command(create_figure)
cli.add_command(create_doc)
cli.add_command(upload_to_zenodo)
cli.add_command(view_doc)
cli.add_command(generate_doc_bibliography)


from setuptools import setup, find_packages

setup(
    packages=find_packages(),
    scripts=[],
    include_package_data=True,
    package_data={
        'figurl_to_html': ['templates/*'],
    },
    install_requires=[
        'kachery-cloud',
        'click',
        'requests'
    ],
    entry_points={
        'console_scripts': [
            'figurl-to-html=figurl_to_html.cli:cli'
        ]
    }
)



import os
from http.server import HTTPServer, SimpleHTTPRequestHandler


def view_doc(*, file_path: str, port: int, allow_origin: str):
    class CORSRequestHandler(SimpleHTTPRequestHandler):
        def end_headers(self):
            self.send_header('Access-Control-Allow-Origin', allow_origin)
            self.send_header('Access-Control-Allow-Methods', 'GET')
            self.send_header('Cache-Control', 'no-store, no-cache, must-revalidate')
            return super(CORSRequestHandler, self).end_headers()

    abs_file_path = os.path.abspath(file_path)
    parent_path = os.path.dirname(abs_file_path)
    fname = os.path.basename(abs_file_path)

    os.chdir(parent_path)
    httpd = HTTPServer(('localhost', port), CORSRequestHandler)
    print(f'https://doc.figurl.org?doc=http://localhost:{port}/{fname}')
    httpd.serve_forever()